@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2008-2023. All Rights Reserved.
+%% Copyright Ericsson AB 2008-2024. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -21,6 +21,7 @@
 %%
 
 -module(ssh_acceptor).
+-moduledoc false.
 
 -include("ssh.hrl").
 
@@ -73,6 +74,9 @@ close(Socket, Options) ->
 acceptor_init(Parent, SystemSup,
               #address{address=Address, port=Port, profile=_Profile},
               Opts) ->
+    ssh_lib:set_label(server,
+                      {acceptor,
+                       list_to_binary(ssh_lib:format_address_port(Address, Port))}),
     AcceptTimeout = ?GET_INTERNAL_OPT(timeout, Opts, ?DEFAULT_TIMEOUT),
     case ?GET_INTERNAL_OPT(lsocket, Opts, undefined) of
         {LSock, SockOwner} ->
@@ -99,7 +103,7 @@ acceptor_init(Parent, SystemSup,
 
         undefined ->
             %% No listening socket (nor fd option) was provided; open a listening socket:
-            case listen(Port, Opts) of
+            case try_listen(Port, Opts, 4) of
                 {ok,LSock} ->
                     proc_lib:init_ack(Parent, {ok, self()}),
                     acceptor_loop(Port, Address, Opts, LSock, AcceptTimeout, SystemSup);
